# -*- coding: utf-8 -*-

""" Cost functions for video calibration, used with scipy. """

# pylint:disable=invalid-name, too-many-positional-arguments

import numpy as np

import sksurgerycalibration.video.video_calibration_metrics as vm
import sksurgerycalibration.video.video_calibration_utils as vu


def stereo_2d_error_for_extrinsics(x_0,
                                   common_object_points,
                                   common_left_image_points,
                                   common_right_image_points,
                                   left_intrinsics,
                                   left_distortion,
                                   right_intrinsics,
                                   right_distortion,
                                   l2r_rmat,
                                   l2r_tvec
                                   ):
    """
    Used with scipy.least_squares for Levenberg-Marquardt optimisation for example.

    Cost function to be used when you are only optimising
    the left camera extrinsic parameters over multiple views,
    and leaving intrinsics, distortion, and left-to-right constant.
    So x_0 should be of length 6 * number_of_frames, corresponding
    to 3 rvec (OpenCV Rodrigues rotation vector) and 3 tvec parameters per frame.

    Computes a vector of residuals between projected image points
    and actual image points, for left and right image.
    """
    rvecs = []
    tvecs = []
    number_of_frames = len(common_object_points)
    for i in range(0, number_of_frames):
        rvec = np.zeros((3, 1))
        rvec[0][0] = x_0[6 * i + 0]
        rvec[1][0] = x_0[6 * i + 1]
        rvec[2][0] = x_0[6 * i + 2]
        tvec = np.zeros((3, 1))
        tvec[0][0] = x_0[6 * i + 3]
        tvec[1][0] = x_0[6 * i + 4]
        tvec[2][0] = x_0[6 * i + 5]
        rvecs.append(rvec)
        tvecs.append(tvec)

    residual = vm.compute_stereo_2d_err(l2r_rmat,
                                        l2r_tvec,
                                        common_object_points,
                                        common_left_image_points,
                                        left_intrinsics,
                                        left_distortion,
                                        common_object_points,
                                        common_right_image_points,
                                        right_intrinsics,
                                        right_distortion,
                                        rvecs,
                                        tvecs,
                                        return_residuals=True
                                        )
    return residual


# pylint:disable=too-many-arguments
def stereo_proj_err_h2e(x_0,
                        common_object_points,
                        common_left_image_points,
                        common_right_image_points,
                        left_intrinsics,
                        left_distortion,
                        right_intrinsics,
                        right_distortion,
                        l2r_rmat,
                        l2r_tvec,
                        device_tracking_array,
                        pattern_tracking_array,
                        left_pattern2marker_matrix=None
                        ):
    """
    Used with scipy.minimize for Powell or Nelder-Mead optimisation for example.

    Cost function to be used when you are optimising the hand-eye matrix.
    If left_pattern2marker_matrix is None, you are also optimising that.

    So, x_0 should be of length 6 or 12. If x_0 is length 6, it corresponds to
    3 rvec and 3 tvec parameters for hand-to-eye. If x_0 is length 12, then the
    first 6 are hand-to-eye, and the second 6 are pattern-to-marker.

    Computes the SSE of projected image points and actual image points for left and right cameras.

    :param x_0:
    :param common_object_points:
    :param common_left_image_points:
    :param common_right_image_points:
    :param left_intrinsics:
    :param left_distortion:
    :param right_intrinsics:
    :param right_distortion:
    :param l2r_rmat:
    :param l2r_tvec:
    :param device_tracking_array:
    :param pattern_tracking_array:
    :param left_pattern2marker_matrix:
    :return: matrix of residuals for Levenberg-Marquardt optimisation.
    """
    rvecs = []
    tvecs = []
    number_of_frames = len(common_object_points)

    h2e_rvec = np.zeros((3, 1))
    h2e_rvec[0][0] = x_0[0]
    h2e_rvec[1][0] = x_0[1]
    h2e_rvec[2][0] = x_0[2]

    h2e_tvec = np.zeros((3, 1))
    h2e_tvec[0][0] = x_0[3]
    h2e_tvec[1][0] = x_0[4]
    h2e_tvec[2][0] = x_0[5]

    h2e = vu.extrinsic_vecs_to_matrix(h2e_rvec, h2e_tvec)

    if left_pattern2marker_matrix is None:

        p2m_rvec = np.zeros((3, 1))
        p2m_rvec[0][0] = x_0[6]
        p2m_rvec[1][0] = x_0[7]
        p2m_rvec[2][0] = x_0[8]

        p2m_tvec = np.zeros((3, 1))
        p2m_tvec[0][0] = x_0[9]
        p2m_tvec[1][0] = x_0[10]
        p2m_tvec[2][0] = x_0[11]

        p2m = vu.extrinsic_vecs_to_matrix(p2m_rvec, p2m_tvec)

    else:

        p2m = left_pattern2marker_matrix

    for i in range(0, number_of_frames):

        p2c = h2e \
              @ np.linalg.inv(device_tracking_array[i]) \
              @ pattern_tracking_array[i] \
              @ p2m

        rvec, tvec = vu.extrinsic_matrix_to_vecs(p2c)

        rvecs.append(rvec)
        tvecs.append(tvec)

    proj, _ = vm.compute_stereo_2d_err(l2r_rmat,
                                       l2r_tvec,
                                       common_object_points,
                                       common_left_image_points,
                                       left_intrinsics,
                                       left_distortion,
                                       common_object_points,
                                       common_right_image_points,
                                       right_intrinsics,
                                       right_distortion,
                                       rvecs,
                                       tvecs
                                       )
    return proj


def stereo_proj_err_h2e_int_dist_l2r(x_0,
                                     common_object_points,
                                     common_left_image_points,
                                     common_right_image_points,
                                     device_tracking_array,
                                     pattern_tracking_array,
                                     left_pattern2marker_matrix
                                     ):
    """
    Similar to stereo_proj_err_h2e, except we are also
    optimising intrinsic and distortion parameters for both cameras,
    and the left-to-right extrinsic parameters.

    So x_0 contains hand-eye rvec, hand-eye tvec, l2r rvec, l2r tvec,
    left intrinsics (fx, fy, cx,cy), left distortion (k1, k2, p1, p2, k3),
    right intrinsics (fx, fy, cx,cy), right distortion (k1, k2, p1, p2, k3).
    So x_0 should be of length 30.

    However, note that: https://doi.org/10.3390/s19122837
    suggests this is a bad idea in general, as you will overfit.
    """
    h2e_rvec = np.zeros((3, 1))
    h2e_rvec[0][0] = x_0[0]
    h2e_rvec[1][0] = x_0[1]
    h2e_rvec[2][0] = x_0[2]

    h2e_tvec = np.zeros((3, 1))
    h2e_tvec[0][0] = x_0[3]
    h2e_tvec[1][0] = x_0[4]
    h2e_tvec[2][0] = x_0[5]

    h2e = vu.extrinsic_vecs_to_matrix(h2e_rvec, h2e_tvec)

    l2r_rvec = np.zeros((3, 1))
    l2r_rvec[0][0] = x_0[6]
    l2r_rvec[1][0] = x_0[7]
    l2r_rvec[2][0] = x_0[8]

    l2r_tvec = np.zeros((3, 1))
    l2r_tvec[0][0] = x_0[9]
    l2r_tvec[1][0] = x_0[10]
    l2r_tvec[2][0] = x_0[11]

    l2r = vu.extrinsic_vecs_to_matrix(l2r_rvec, l2r_tvec)

    left_intrinsics = np.zeros((3, 3))
    left_intrinsics[0][0] = x_0[12]
    left_intrinsics[1][1] = x_0[13]
    left_intrinsics[0][2] = x_0[14]
    left_intrinsics[1][2] = x_0[15]

    left_distortion = np.zeros((1, 5))
    left_distortion[0][0] = x_0[16]
    left_distortion[0][1] = x_0[17]
    left_distortion[0][2] = x_0[18]
    left_distortion[0][3] = x_0[19]
    left_distortion[0][4] = x_0[20]

    right_intrinsics = np.zeros((3, 3))
    right_intrinsics[0][0] = x_0[21]
    right_intrinsics[1][1] = x_0[22]
    right_intrinsics[0][2] = x_0[23]
    right_intrinsics[1][2] = x_0[24]

    right_distortion = np.zeros((1, 5))
    right_distortion[0][0] = x_0[25]
    right_distortion[0][1] = x_0[26]
    right_distortion[0][2] = x_0[27]
    right_distortion[0][3] = x_0[28]
    right_distortion[0][4] = x_0[29]

    rvecs = []
    tvecs = []
    number_of_frames = len(common_object_points)

    for i in range(0, number_of_frames):

        p2c = h2e \
              @ np.linalg.inv(device_tracking_array[i]) \
              @ pattern_tracking_array[i] \
              @ left_pattern2marker_matrix

        rvec, tvec = vu.extrinsic_matrix_to_vecs(p2c)

        rvecs.append(rvec)
        tvecs.append(tvec)

    proj, _ = vm.compute_stereo_2d_err(l2r[0:3, 0:3],
                                       l2r[0:3, 3],
                                       common_object_points,
                                       common_left_image_points,
                                       left_intrinsics,
                                       left_distortion,
                                       common_object_points,
                                       common_right_image_points,
                                       right_intrinsics,
                                       right_distortion,
                                       rvecs,
                                       tvecs
                                       )
    return proj


def mono_proj_err_h2e(x_0,
                      object_points,
                      image_points,
                      intrinsics,
                      distortion,
                      pattern_tracking,
                      device_tracking,
                      pattern2marker_matrix
                      ):
    """
    Used with scipy.minimize for Powell or Nelder-Mead optimisation for example.

    Cost function to be used when you are only optimising the hand-eye matrix.
    So, x_0 should be of length 6, corresponding to 3 rvec and 3 tvec parameters.

    Computes the SSE of projected image points and actual image points. Computes
    the pattern-to-camera matrix for each frame, using pattern-to-marker,
    marker-to-world (calibration pattern tracking), the inverse of the
    device-to-world (device tracking, e.g. laparoscope), and hand-to-eye.
    """
    assert len(x_0) == 6

    rvec = np.zeros((3, 1))
    rvec[0][0] = x_0[0]
    rvec[1][0] = x_0[1]
    rvec[2][0] = x_0[2]

    tvec = np.zeros((3, 1))
    tvec[0][0] = x_0[3]
    tvec[1][0] = x_0[4]
    tvec[2][0] = x_0[5]

    h2e = vu.extrinsic_vecs_to_matrix(rvec, tvec)

    number_of_frames = len(object_points)
    rvecs = []
    tvecs = []

    # Computes pattern2camera for each pose
    for i in range(0, number_of_frames):

        p2c = h2e @ np.linalg.inv(device_tracking[i]) @ \
              pattern_tracking[i] @ pattern2marker_matrix

        rvec, tvec = vu.extrinsic_matrix_to_vecs(p2c)

        rvecs.append(rvec)
        tvecs.append(tvec)

    proj, _ = vm.compute_mono_2d_err(object_points,
                                     image_points,
                                     rvecs,
                                     tvecs,
                                     intrinsics,
                                     distortion,
                                     return_residuals=False)
    return proj


def mono_proj_err_p2m_h2e(x_0,
                          object_points,
                          image_points,
                          intrinsics,
                          distortion,
                          pattern_tracking,
                          device_tracking
                          ):
    """
    Similar to mono_proj_err_h2e, except x_0 should be
    of length 12, corresponding to pattern2marker and hand2eye.

    So 3 pattern2marker rvec, 3 pattern2marker tvec,
    3 hand2eye rvec, 3 hand2eye tvec.
    """
    assert len(x_0) == 12

    rvec = np.zeros((3, 1))
    rvec[0][0] = x_0[0]
    rvec[1][0] = x_0[1]
    rvec[2][0] = x_0[2]

    tvec = np.zeros((3, 1))
    tvec[0][0] = x_0[3]
    tvec[1][0] = x_0[4]
    tvec[2][0] = x_0[5]

    p2m = vu.extrinsic_vecs_to_matrix(rvec, tvec)

    rvec[0][0] = x_0[6]
    rvec[1][0] = x_0[7]
    rvec[2][0] = x_0[8]

    tvec[0][0] = x_0[9]
    tvec[1][0] = x_0[10]
    tvec[2][0] = x_0[11]

    h2e = vu.extrinsic_vecs_to_matrix(rvec, tvec)

    number_of_frames = len(object_points)
    rvecs = []
    tvecs = []

    # Computes pattern2camera for each pose
    for i in range(0, number_of_frames):

        p2c = h2e @ np.linalg.inv(device_tracking[i])\
              @ pattern_tracking[i] @ p2m

        rvec, tvec = vu.extrinsic_matrix_to_vecs(p2c)

        rvecs.append(rvec)
        tvecs.append(tvec)

    proj, _ = vm.compute_mono_2d_err(object_points,
                                     image_points,
                                     rvecs,
                                     tvecs,
                                     intrinsics,
                                     distortion,
                                     return_residuals=False)
    return proj


def mono_proj_err_h2e_g2w(x_0,
                          object_points,
                          image_points,
                          intrinsics,
                          distortion,
                          device_tracking
                          ):
    """
    Similar to mono_proj_err_p2m_h2e, except, as is the case
    in robot-world-hand-eye calibration, we assume we have
    a stationary untracked pattern, and we are estimating
    both the hand2eye and grid2world (i.e. pattern2world).

    So, x_0 should be of length 12, corresponding to
    3 hand2eye rvec, 3 hand2eye tvec,
    3 grid2world rvec, 3 grid2world tvec.
    """
    assert len(x_0) == 12

    rvec = np.zeros((3, 1))
    rvec[0][0] = x_0[0]
    rvec[1][0] = x_0[1]
    rvec[2][0] = x_0[2]

    tvec = np.zeros((3, 1))
    tvec[0][0] = x_0[3]
    tvec[1][0] = x_0[4]
    tvec[2][0] = x_0[5]

    h2e = vu.extrinsic_vecs_to_matrix(rvec, tvec)

    rvec[0][0] = x_0[6]
    rvec[1][0] = x_0[7]
    rvec[2][0] = x_0[8]

    tvec[0][0] = x_0[9]
    tvec[1][0] = x_0[10]
    tvec[2][0] = x_0[11]

    g2w = vu.extrinsic_vecs_to_matrix(rvec, tvec)

    number_of_frames = len(object_points)
    rvecs = []
    tvecs = []

    # Computes pattern2camera for each pose
    for i in range(0, number_of_frames):

        p2c = h2e @ np.linalg.inv(device_tracking[i]) @ g2w

        rvec, tvec = vu.extrinsic_matrix_to_vecs(p2c)

        rvecs.append(rvec)
        tvecs.append(tvec)

    proj, _ = vm.compute_mono_2d_err(object_points,
                                     image_points,
                                     rvecs,
                                     tvecs,
                                     intrinsics,
                                     distortion,
                                     return_residuals=False)
    return proj


def mono_proj_err_h2e_int_dist(x_0,
                               object_points,
                               image_points,
                               device_tracking,
                               pattern_tracking,
                               pattern2marker_matrix
                               ):
    """
    Similar to mono_proj_err_h2e, except we are also optimising
    intrinsic and distortion parameters. So, x_0 contains
    rvec and tvec for hand2eye (6 DOF), then 4 intrinsics
    (fx, fy, cx, cy), then 5 distortion parameters
    (k1, k2, p1, p2, k3).

    However, note that: https://doi.org/10.3390/s19122837
    suggests this is a bad idea in general, as you will overfit.
    """
    assert len(x_0) == 15

    rvec = np.zeros((3, 1))
    rvec[0][0] = x_0[0]
    rvec[1][0] = x_0[1]
    rvec[2][0] = x_0[2]

    tvec = np.zeros((3, 1))
    tvec[0][0] = x_0[3]
    tvec[1][0] = x_0[4]
    tvec[2][0] = x_0[5]

    h2e = vu.extrinsic_vecs_to_matrix(rvec, tvec)

    intrinsics = np.zeros((3, 3))
    intrinsics[0][0] = x_0[6]
    intrinsics[1][1] = x_0[7]
    intrinsics[0][2] = x_0[8]
    intrinsics[1][2] = x_0[9]

    distortion = np.zeros((1, 5))
    distortion[0][0] = x_0[10]
    distortion[0][1] = x_0[11]
    distortion[0][2] = x_0[12]
    distortion[0][3] = x_0[13]
    distortion[0][4] = x_0[14]

    number_of_frames = len(object_points)
    rvecs = []
    tvecs = []

    # Computes pattern2camera for each pose
    for i in range(0, number_of_frames):

        p2c = h2e @ np.linalg.inv(device_tracking[i]) @ \
              pattern_tracking[i] @ pattern2marker_matrix

        rvec, tvec = vu.extrinsic_matrix_to_vecs(p2c)

        rvecs.append(rvec)
        tvecs.append(tvec)

    proj, _ = vm.compute_mono_2d_err(object_points,
                                     image_points,
                                     rvecs,
                                     tvecs,
                                     intrinsics,
                                     distortion)
    return proj



# -*- coding: utf-8 -*-

""" Various routines for Hand-Eye calibration. """

# pylint:disable=invalid-name,consider-using-enumerate

import logging
from typing import List, Tuple

import cv2
import numpy as np
import sksurgerycore.transforms.matrix as mu
from scipy.optimize import least_squares

import sksurgerycalibration.video.quaternion_utils as qu
import sksurgerycalibration.video.video_calibration_utils as vu

LOGGER = logging.getLogger(__name__)


def set_model2hand_arrays(calibration_tracking_array: List,
                          device_tracking_array: List,
                          use_quaternions=False) \
                            -> Tuple[np.ndarray, np.ndarray]:
    """
    Guofang Xiao's method. Set the model-to-hand quaternion
    and translation arrays from tracking data.

    :param calibration_tracking_array: Array of tracking data for
    calibration target
    :type calibration_tracking_array: List of tracking data
    :param device_tracking_array: Array of tracking data for
    device (e.g. camera)
    :type device_tracking_array: List of tracking data
    :param use_quaternions: If True input should be quaternions
    :type device_tracking_array: bool
    :return: quaternion model to hand array and translation model to hand
    array
    :rtype: np.ndarray, np.ndarray
    """

    if len(calibration_tracking_array) != len(device_tracking_array):
        raise ValueError('Calibration target and device tracking array \
                            should be the same size')

    number_of_frames = len(calibration_tracking_array)

    quat_model2hand_array = np.zeros((number_of_frames, 4))
    trans_model2hand_array = np.zeros((number_of_frames, 3))

    for i in range(number_of_frames):
        if use_quaternions:
            quat_model = \
                calibration_tracking_array[i][0, 0:4]
            trans_model = \
                calibration_tracking_array[i][0, 4:7]
            quat_hand = \
                device_tracking_array[i][0, 0:4]
            trans_hand = \
                device_tracking_array[i][0, 4:7]

            quat_model2hand = qu.quat_multiply(qu.quat_conjugate(quat_hand),
                                            quat_model)
            q_translation = np.append([0], (trans_model - trans_hand))
            quat_hand_conj = qu.quat_conjugate(quat_hand)

            trans_model2hand = qu.quat_multiply(quat_hand_conj,
                                                qu.quat_multiply(q_translation,
                                                                 quat_hand))
            trans_model2hand = trans_model2hand[1:4]
        else:
            tracking_model = calibration_tracking_array[i]
            tracking_hand = device_tracking_array[i]
            model_to_hand = np.linalg.inv(tracking_hand) @ tracking_model

            quat_model2hand = qu.rotm_to_quat(model_to_hand[0:3, 0:3])
            trans_model2hand = model_to_hand[0:3, 3]

        quat_model2hand_array[i] = quat_model2hand
        trans_model2hand_array[i] = trans_model2hand

    quat_model2hand_array = \
        qu.to_one_hemisphere(quat_model2hand_array)

    return quat_model2hand_array, trans_model2hand_array


def _gx_solve_2quaternions(qx, q_A, q_B):
    """
    Guofang Xiao's method. Provide cost function for least-square optimisation
    to solve quaternions qx1 and qx2 in: q_B = qx1 * q_A * qx2

    :param qx: 1x8 vector of 2 quaternions
    :param q_A: Nx4 matrices of N quaternions
    :param q_B: Nx4 matrices of N quaternions
    :return:
    """
    if np.shape(q_A) != np.shape(q_B):
        raise AssertionError("q_A and q_B must have the same shape.")

    number_of_frames = np.shape(q_A)[0]

    f = []

    qx_1 = qx[0:4]
    qx_2 = qx[4:8]

    qx_1 = qx_1/np.linalg.norm(qx_1)
    qx_2 = qx_2/np.linalg.norm(qx_2)

    for i in range(0, number_of_frames):
        f.append(qu.quat_multiply(qu.quat_multiply(qx_1, q_A[i]), qx_2)
                 - q_B[i])

    f = np.ndarray.flatten(np.asarray(f))

    return f


def _gx_solve_2translations(q_handeye,
                            quat_model2hand_array,
                            trans_model2hand_array,
                            trans_extrinsics_array):
    """
    Guofang Xiao's method. Solve translations t_handeye and t_pattern2marker.
    translations = [t_handeye t_pattern2marker]

    :param q_handeye: 1x4 matrix quaternion
    :param quat_model2hand_array: Nx4 matrix of N quaternions
    :param trans_model2hand_array: Nx3 matrices of N translations
    :param trans_extrinsics_array: Nx3 matrices of extrinsic translation vectors
    :return:
    """
    if np.shape(quat_model2hand_array)[0] \
            != np.shape(trans_model2hand_array)[0]:
        raise AssertionError("Wrong dimensions in solve_2translations().")
    if np.shape(trans_model2hand_array) != np.shape(trans_extrinsics_array):
        raise AssertionError("Wrong dimensions in solve_2translations().")

    number_of_frames = np.shape(quat_model2hand_array)[0]

    A = np.zeros((3 * number_of_frames, 6))
    B = np.zeros((3 * number_of_frames, 1))

    for i in range(0, number_of_frames):
        qHEqMH = qu.quat_multiply(q_handeye, quat_model2hand_array[i])
        rHErMH = qu.quat_to_rotm(qHEqMH)
        rHE = qu.quat_to_rotm(q_handeye)

        A[i*3:i*3+3, :] = np.hstack([np.identity(3), rHErMH])
        b_i = np.transpose(trans_extrinsics_array[i, :])\
            - np.matmul(rHE, np.transpose(trans_model2hand_array[i, :]))
        B[i*3:i*3+3] = np.reshape(b_i, (3, 1))

    A_T = np.transpose(A)
    translations = np.linalg.inv(A_T @ A) @ A_T @ B
    translations = np.transpose(translations)

    return translations.flatten()


def _gx_handeye_optimisation(
        quat_extrinsics_array: np.ndarray,
        trans_extrinsics_array: np.ndarray,
        quat_model2hand_array: np.ndarray,
        trans_model2hand_array: np.ndarray) ->  \
        Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """
    Guofang Xiao's method. Solve handeye and pattern-to-marker transformations.

    :param quat_extrinsics_array: An array of quaternions representing the
        rotations of the camera extrinsics.
    :type quat_extrinsics_array: np.ndarray
    :param trans_extrinsics_array:  Array of the translation vectors of the
        camera extrinsics.
    :type trans_extrinsics_array: np.ndarray
    :param quat_model2hand_array: An array of model to hand quaternions.
        :type quat_model2hand_array: np.ndarray
    :param quat_model2hand_array: An array of model to hand
        translations arrays.
        :type quat_model2hand_array: np.ndarray
    :return: rotations in quaternion form and translations of the handeye
        pattern-to-marker transformation.
    :rtype: np.ndarray
    """

    # self.quat_model2hand_array should already be in one hemisphere.

    quat_extrinsics_array = qu.to_one_hemisphere(quat_extrinsics_array)

    qx_0 = [1, 0, 0, 0, 1, 0, 0, 0]
    lb = [-1, -1, -1, -1, -1, -1, -1, -1]
    ub = [1, 1, 1, 1, 1, 1, 1, 1]

    op_result = least_squares(_gx_solve_2quaternions, qx_0,
                              bounds=(lb, ub),
                              args=(quat_model2hand_array,
                                    quat_extrinsics_array)
                              )

    q_handeye = op_result.x[0:4]
    q_pattern2marker = op_result.x[4:8]

    q_handeye = q_handeye / np.linalg.norm(q_handeye)
    q_pattern2marker = q_pattern2marker / np.linalg.norm(q_pattern2marker)

    translations = _gx_solve_2translations(q_handeye,
                                           quat_model2hand_array,
                                           trans_model2hand_array,
                                           trans_extrinsics_array)

    t_handeye = translations[0:3]
    t_pattern2marker = translations[3:6]

    return q_handeye, t_handeye, q_pattern2marker, t_pattern2marker


def guofang_xiao_handeye_calibration(rvecs: List[np.ndarray],
                                     tvecs: List[np.ndarray],
                                     quat_model2hand_array: np.ndarray,
                                     trans_model2hand_array: np.ndarray) \
                                     -> Tuple[np.ndarray, np.ndarray]:
    """
    Guofang Xiao's method. Solve for the hand-eye transformation,
    as well as the transformation from the pattern to the tracking markers
    on the calibration object.

    This solves rotation and then translation in two steps, using least-squares
    optimisation. It uses quaternions to represent rotations.

    This method, developed for the SmartLiver project, assumes both
    device (laparoscope), and the calibration object are tracked. While
    both calibration object and laparoscope are tracked,
    so you could move things freely, we also developed a calibration
    rig. So, in general, the laparoscope is stationary on the rig and the calibration
    object is moved. This is mostly to reduce image blur and jitter
    caused by hand movements. See Dowrick  et al., 2023,
    https://doi.org/10.1002/mp.16310.

    :param rvecs: Array of rotation vectors, from OpenCV, camera extrinsics
    (pattern to camera transform)
    :type rvecs: List[np.ndarray]
    :param tvecs: Array of translation vectors, from OpenCV, camera extrinsics
    (pattern to camera transform)
    :type tvecs: List[np.ndarray]
    :param quat_model2hand_array: Array of quaternions representing rotational
    part of marker-to-hand.
    :type quat_model2hand_array: np.ndarray
    :param trans_model2hand_array: Array of quaternions representing
    translational part of marker-to-hand.
    :type trans_model2hand_array: np.ndarray
    :return: two 4x4 matrices as np.ndarray, representing handeye_matrix,
    pattern2marker_matrix
    :rtype: np.ndarray, np.ndarray
    """

    number_of_frames = len(rvecs)
    quat_extrinsics_array = np.zeros((number_of_frames, 4))

    for i in range(number_of_frames):
        quat_extrinsics_array[i] = qu.rvec_to_quat(rvecs[i].flatten())

    trans_extrinsics_array = np.reshape(tvecs, (number_of_frames, 3))

    q_handeye, t_handeye, q_pattern2marker, t_pattern2marker = \
        _gx_handeye_optimisation(quat_extrinsics_array, trans_extrinsics_array,
                                 quat_model2hand_array, trans_model2hand_array)

    handeye_matrix = np.eye(4)
    handeye_matrix[0:3, 0:3] = qu.quat_to_rotm(q_handeye)
    handeye_matrix[0:3, 3] = t_handeye

    pattern2marker_matrix = np.eye(4)
    pattern2marker_matrix[0:3, 0:3] = qu.quat_to_rotm(q_pattern2marker)
    pattern2marker_matrix[0:3, 3] = t_pattern2marker

    return handeye_matrix, pattern2marker_matrix


def calibrate_hand_eye_using_stationary_pattern(
        camera_rvecs: List[np.ndarray],
        camera_tvecs: List[np.ndarray],
        tracking_matrices: List[np.ndarray],
        method=cv2.CALIB_HAND_EYE_TSAI,
        invert_camera=False
        ):
    """
    Hand-eye calibration using standard OpenCV methods. This method assumes
    a single set of tracking data, so it is useful for a stationary, untracked
    calibration pattern, and a tracked video device, e.g. laparoscope.

    Please do read Ali et al., 2019, https://doi.org/10.3390/s19122837.

    OpenCV implements Tsai, Horaud and Park methods, which compute
    rotation then translation, as well as Andref and Daniilidis methods.
    See cv.calibrateHandEye and enum cv::HandEyeCalibrationMethod.
    Internally this method calls cv.calibrateHandEye.

    On average, in Ali's evaluation, Horaud might be a good default,
    but I've left it with Tsai's as the most well known (perhaps?).

    :param camera_rvecs: list of rvecs that we get from OpenCV camera
    extrinsics, pattern_to_camera.
    :param camera_tvecs: list of tvecs that we get from OpenCV camera
    extrinsics, pattern_to_camera.
    :param tracking_matrices: list of tracking matrices for the tracked
    device, marker_to_tracker.
    :param method: Choice of OpenCV Hand-Eye method.
    :param invert_camera: if True, we invert camera matrices before
    hand-eye calibration.
    :return hand-eye transform as 4x4 matrix.
    """
    if len(camera_rvecs) != len(camera_tvecs):
        raise ValueError("Camera rotation and translation vector "
                         "lists must be the same length.")

    if len(camera_tvecs) != len(tracking_matrices):
        raise ValueError("The number of camera extrinsic transforms must "
                         "equal the number of tracking matrices.")

    if len(camera_rvecs) < 3:
        raise ValueError("You must have at least 3 views, include movements "
                         "around at least 2 different rotation axes.")

    # Convert tracking matrices to rvecs/tvecs for OpenCV.
    tracking_rvecs = []
    tracking_tvecs = []
    for i in range(len(camera_rvecs)):
        rvecs, tvecs = vu.extrinsic_matrix_to_vecs(tracking_matrices[i])
        tracking_rvecs.append(rvecs)
        tracking_tvecs.append(tvecs)

    cam_rvecs = []
    cam_tvecs = []
    for i in range(len(camera_rvecs)):
        mat = vu.extrinsic_vecs_to_matrix(camera_rvecs[i], camera_tvecs[i])
        if invert_camera:
            mat = np.linalg.inv(mat)
        rvecs, tvecs = vu.extrinsic_matrix_to_vecs(mat)
        cam_rvecs.append(rvecs)
        cam_tvecs.append(tvecs)

    # OpenCV outputs eye-to-hand.
    e2h_rmat, e2h_tvec = cv2.calibrateHandEye(tracking_rvecs,
                                              tracking_tvecs,
                                              cam_rvecs,
                                              cam_tvecs,
                                              method=method
                                              )

    e2h = mu.construct_rigid_transformation(e2h_rmat, e2h_tvec)
    h2e = np.linalg.inv(e2h)

    return h2e


# pylint:disable=too-many-positional-arguments
def calibrate_hand_eye_using_tracked_pattern(
        camera_rvecs: List[np.ndarray],
        camera_tvecs: List[np.ndarray],
        device_tracking_matrices: List[np.ndarray],
        pattern_tracking_matrices: List[np.ndarray],
        method=cv2.CALIB_HAND_EYE_TSAI,
        invert_camera=False
        ):
    """
    See comments for calibrate_hand_eye_using_stationary_pattern.

    Here, we also have pattern_tracking_matrices. So, device_tracking_matrices
    and pattern_tracking_matrices are combined to give marker-to-hand
    transformation, and then calibrate_hand_eye_using_stationary_pattern called.

    You will notice, this method is NOT calculating pattern_to_marker. (i.e.
    the transformation between a calibration pattern, and the tracking marker
    attached to it). So, essentially, this method is simply treating the
    calibration pattern as the reference for the tracker space.
    This means you can hold the calibration pattern, and laparoscope,
    and move both of them around independently, but the maths is solved
    by assuming the calibration pattern is stationary, and the laparoscope
    is moving around the calibration pattern.
    """
    if len(device_tracking_matrices) != len(pattern_tracking_matrices):
        raise ValueError("The number of device tracking matrices must "
                         "equal the number of pattern tracking matrices.")

    if len(device_tracking_matrices) < 3:
        raise ValueError("You must have at least 3 tracking matrices")

    for i in range(len(pattern_tracking_matrices)):
        if pattern_tracking_matrices[i] is None:
            raise ValueError("Pattern tracking matrix:"
                             + str(i) + str(" is None."))

    # Essentially, as we have a tracked calibration object,
    # the calibration object becomes the reference for the tracker space.
    tracking_matrices = []
    for i in range(len(device_tracking_matrices)):
        mat = np.linalg.inv(pattern_tracking_matrices[i]) \
              @ device_tracking_matrices[i]
        tracking_matrices.append(mat)

    h2e = calibrate_hand_eye_using_stationary_pattern(camera_rvecs,
                                                      camera_tvecs,
                                                      tracking_matrices,
                                                      method,
                                                      invert_camera)
    return h2e


def calibrate_pattern_to_tracking_marker(camera_rvecs: List[np.ndarray],
                                         camera_tvecs: List[np.ndarray],
                                         tracking_matrices: List[np.ndarray],
                                         method=cv2.CALIB_HAND_EYE_TSAI
                                         ):
    """
    This is really a convenience method that is analagous to
    calibrate_hand_eye_using_stationary_pattern. i.e. you have a stationary
    camera but a moving calibration pattern.

    So, here, you are JUST calculating pattern_to_marker transformation.

    :param camera_rvecs: list of rvecs that we get from OpenCV camera
    extrinsics, pattern_to_camera.
    :param camera_tvecs: list of tvecs that we get from OpenCV camera
    extrinsics, pattern_to_camera.
    :param tracking_matrices: list of pattern tracking matrices for the tracked
    calibration pattern, marker_to_tracker.
    :param method: Choice of OpenCV Hand-Eye method.
    :return pattern_to_marker
    """
    h2e = calibrate_hand_eye_using_stationary_pattern(camera_rvecs,
                                                      camera_tvecs,
                                                      tracking_matrices,
                                                      method,
                                                      invert_camera=True)

    p2m = np.linalg.inv(h2e)

    return p2m


def calibrate_hand_eye_and_pattern_to_marker(
        camera_rvecs: List[np.ndarray],
        camera_tvecs: List[np.ndarray],
        device_tracking_matrices: List[np.ndarray],
        pattern_tracking_matrices: List[np.ndarray],
        method=cv2.CALIB_ROBOT_WORLD_HAND_EYE_SHAH
        ):
    """
    Hand-eye calibration using standard OpenCV methods. This method assumes
    you are tracking both the device that needs hand-eye calibration, and
    the calibration pattern, and internally calls cv.calibrateRobotWorldHandEye.

    Again, please do read Ali et al., 2019, https://doi.org/10.3390/s19122837.

    OpenCV already implements Shah and Li's methods. See cv.calibrateRobotWorldHandEye.

    :param camera_rvecs: list of rvecs that we get from OpenCV camera
    extrinsics, pattern_to_camera.
    :param camera_tvecs: list of tvecs that we get from OpenCV camera
    extrinsics, pattern_to_camera.
    :param device_tracking_matrices: list of tracking matrices for the
    tracked device, e.g. laparoscope, marker_to_tracker.
    :param pattern_tracking_matrices: list of tracking matrices for the
    calibration object, marker_to_tracker
    :param method: Choice of OpenCV RobotWorldHandEye method.
    :return hand-eye, pattern-to-marker transforms as 4x4 matrices
    """
    if len(camera_rvecs) != len(camera_tvecs):
        raise ValueError("Camera rotation and translation vector lists "
                         "must be the same length.")

    if len(camera_tvecs) != len(device_tracking_matrices):
        raise ValueError("The number of camera extrinsic transforms must "
                         "equal the number of device tracking matrices.")

    if len(camera_tvecs) != len(pattern_tracking_matrices):
        raise ValueError("The number of camera extrinsic transforms must "
                         "equal the number of pattern tracking matrices.")

    if len(camera_rvecs) < 3:
        raise ValueError("You must have at least 3 views, include movements "
                         "around at least 2 different rotation axes.")

    # Convert tracking matrices to rvecs/tvecs for OpenCV.
    # Look at OpenCV documentation for cv.calibrationRobotWorldHandEye
    # Gripper = tracking marker on calibration pattern
    # Robot base = tracking marker on video device, eg. laparoscope
    # World = video camera on laparoscope, assumed to be stationary
    # So, we move the calibration pattern, not the video device.

    base2gripper_rvecs = []
    base2gripper_tvecs = []
    for i in range(len(camera_rvecs)):
        b2g = np.linalg.inv(pattern_tracking_matrices[i]) \
              @ device_tracking_matrices[i]
        rvecs, tvecs = vu.extrinsic_matrix_to_vecs(b2g)
        base2gripper_rvecs.append(rvecs)
        base2gripper_tvecs.append(tvecs)

    cam_rvecs = []
    cam_tvecs = []
    for i in range(len(camera_rvecs)):
        mat = vu.extrinsic_vecs_to_matrix(camera_rvecs[i], camera_tvecs[i])
        mat = np.linalg.inv(mat)
        rvecs, tvecs = vu.extrinsic_matrix_to_vecs(mat)
        cam_rvecs.append(rvecs)
        cam_tvecs.append(tvecs)

    b2w_rmat, b2w_t, g2c_rmat, g2c_t = \
        cv2.calibrateRobotWorldHandEye(cam_rvecs,
                                       cam_tvecs,
                                       base2gripper_rvecs,
                                       base2gripper_tvecs,
                                       method=method)

    h2e = mu.construct_rigid_transformation(b2w_rmat, b2w_t)
    m2p = mu.construct_rigid_transformation(g2c_rmat, g2c_t)
    p2m = np.linalg.inv(m2p)

    return h2e, p2m


def calibrate_hand_eye_and_grid_to_world(
        camera_rvecs: List[np.ndarray],
        camera_tvecs: List[np.ndarray],
        device_tracking_matrices: List[np.ndarray],
        method=cv2.CALIB_ROBOT_WORLD_HAND_EYE_SHAH
        ):
    """
    Similar to calibrate_hand_eye_and_pattern_to_marker, except
    here, the calibration pattern is untracked, and assumed to be
    stationary.

    :param camera_rvecs: list of rvecs that we get from OpenCV camera
    extrinsics, pattern_to_camera.
    :param camera_tvecs: list of tvecs that we get from OpenCV camera
    extrinsics, pattern_to_camera.
    :param device_tracking_matrices: list of tracking matrices for the
    tracked device, e.g. laparoscope, marker_to_tracker.
    :param method: Choice of OpenCV RobotWorldHandEye method.
    :return hand-eye, grid-to-world transforms as 4x4 matrices
    """
    if len(camera_rvecs) != len(camera_tvecs):
        raise ValueError("Camera rotation and translation vector "
                         "lists must be the same length.")

    if len(camera_tvecs) != len(device_tracking_matrices):
        raise ValueError("The number of camera extrinsic transforms "
                         "must equal the number of device tracking matrices.")

    if len(camera_rvecs) < 3:
        raise ValueError("You must have at least 3 views, include "
                         "movements around at least 2 different rotation axes.")

    base2gripper_rvecs = []
    base2gripper_tvecs = []
    for i in range(len(device_tracking_matrices)):
        b2g = np.linalg.inv(device_tracking_matrices[i])
        rvec, tvec = vu.extrinsic_matrix_to_vecs(b2g)
        base2gripper_rvecs.append(rvec)
        base2gripper_tvecs.append(tvec)

    b2w_rmat, b2w_t, g2c_rmat, g2c_t = \
        cv2.calibrateRobotWorldHandEye(camera_rvecs,
                                       camera_tvecs,
                                       base2gripper_rvecs,
                                       base2gripper_tvecs,
                                       method=method)

    h2e = mu.construct_rigid_transformation(g2c_rmat, g2c_t)
    b2w = mu.construct_rigid_transformation(b2w_rmat, b2w_t)
    w2b = np.linalg.inv(b2w)

    return h2e, w2b



# coding=utf-8

"""scikit-surgerynditracker tests"""

# Pytest style

def test_requirements_vs_setup():
    """
    Test that the requirements.txt matches setup.py
    """

   # setup=open("setup.py","r")
   # setuptxt = setup.read()
   # setup.close()

    with open("setup.py", "r", encoding='us-ascii') as file_to_read:
        searchlines = file_to_read.readlines()

    install_line = -1
    for i, line in enumerate(searchlines):
        if "install_requires" in line:
            install_line = i
            break

    end_line = -1
    for i, line in enumerate(searchlines[install_line:]):
        if "]" in line:
            end_line = install_line + i
            break

    setup_reqs = []

    for line in searchlines[install_line + 1: end_line]:
        req = line.replace(',', '').replace("'", "")
        req = req.replace(' ', '').replace('\n', '')
        setup_reqs.append(req)


    with open("requirements.txt", "r", encoding='us-ascii') as file_to_read:
        searchlines = file_to_read.readlines()

    reqs = []
    for line in searchlines:
        if "#" not in line:
            reqs.append(line.replace('\n', ''))

    for setup_req in setup_reqs:
        assert setup_req in reqs

    for req in reqs:
        assert req in setup_reqs



# pylint: disable=line-too-long, missing-module-docstring, invalid-name, missing-function-docstring

import cv2
import sksurgeryimage.calibration.charuco_plus_chessboard_point_detector \
    as charuco_pd
import sksurgerycalibration.video.video_calibration_driver_stereo as sc
import tests.video.test_load_calib_utils as lcu


def get_calib_driver(calib_dir: str):
    """ Create left/right charuco point detectors and load calibration images from directory. """
    minimum_points = 50
    number_of_squares = [19, 26]
    square_tag_sizes = [5, 4]
    number_of_chessboard_squares = [9, 14]
    chessboard_square_size = 3
    chessboard_id_offset = 500

    left_pd = \
        charuco_pd.CharucoPlusChessboardPointDetector(
            dictionary=cv2.aruco.getPredefinedDictionary(cv2.aruco.DICT_4X4_250),
            minimum_number_of_points=minimum_points,
            number_of_charuco_squares=number_of_squares,
            size_of_charuco_squares=square_tag_sizes,
            number_of_chessboard_squares=number_of_chessboard_squares,
            chessboard_square_size=chessboard_square_size,
            chessboard_id_offset=chessboard_id_offset
        )

    right_pd = \
        charuco_pd.CharucoPlusChessboardPointDetector(
            dictionary=cv2.aruco.getPredefinedDictionary(cv2.aruco.DICT_4X4_250),
            minimum_number_of_points=minimum_points,
            number_of_charuco_squares=number_of_squares,
            size_of_charuco_squares=square_tag_sizes,
            number_of_chessboard_squares=number_of_chessboard_squares,
            chessboard_square_size=chessboard_square_size,
            chessboard_id_offset=chessboard_id_offset
        )

    calibration_driver = sc.StereoVideoCalibrationDriver(left_pd,
                                                         right_pd,
                                                         minimum_points)

    for i in range(3):
        l_img, r_img, chessboard, scope = lcu.get_calib_data(calib_dir, i)
        calibration_driver.grab_data(l_img, r_img, scope, chessboard)

    return calibration_driver


def test_charuco_dataset_A():

    calib_dir = 'tests/data/precalib/precalib_base_data'
    calib_driver = get_calib_driver(calib_dir)

    stereo_reproj_err, stereo_recon_err, _ = \
        calib_driver.calibrate()

    tracked_reproj_err, tracked_recon_err, _ = \
        calib_driver.handeye_calibration(use_opencv=False)

    print(stereo_reproj_err, stereo_recon_err, tracked_reproj_err, tracked_recon_err)
    assert stereo_reproj_err < 0.5
    assert stereo_recon_err < 1.5
    assert tracked_reproj_err < 0.5
    assert tracked_recon_err < 1.5


def test_charuco_dataset_B():

    calib_dir = 'tests/data/precalib/data_moved_scope'
    calib_driver = get_calib_driver(calib_dir)

    stereo_reproj_err, stereo_recon_err, _ = \
        calib_driver.calibrate()

    tracked_reproj_err, tracked_recon_err, _ = \
        calib_driver.handeye_calibration(use_opencv=False)

    print(stereo_reproj_err, stereo_recon_err, tracked_reproj_err, tracked_recon_err)
    assert stereo_reproj_err < 0.5
    assert stereo_recon_err < 2.0
    assert tracked_reproj_err < 0.5
    assert tracked_recon_err < 1.0



# -*- coding: utf-8 -*-

""" Video Calibration functions, that wrap OpenCV functions mainly. """

# pylint:disable=too-many-lines, too-many-positional-arguments

import copy
import logging
from typing import List, Tuple

import cv2
import numpy as np
from scipy.optimize import least_squares
from scipy.optimize import minimize

import sksurgerycore.transforms.matrix as skcm
import sksurgerycore.utilities.validate_matrix as cvm
import sksurgerycalibration.video.video_calibration_cost_functions as vcf
import sksurgerycalibration.video.video_calibration_hand_eye as he
import sksurgerycalibration.video.video_calibration_metrics as vm
import sksurgerycalibration.video.video_calibration_utils as vu

LOGGER = logging.getLogger(__name__)


def mono_video_calibration(object_points: List,
                           image_points: List,
                           image_size: Tuple[int, int],
                           camera_matrix: np.ndarray,
                           distortion_coefficients: np.ndarray,
                           flags=0):
    """
    Calibrates a video camera using Zhang's 2000 method, as implemented in
    OpenCV. We wrap it here, so we have a place to add extra validation code,
    and a space for documentation. The aim is to check everything before
    we pass it to OpenCV, and raise Exceptions consistently for any error
    we can detect before we pass it to OpenCV, as OpenCV may die
    without throwing exceptions.

      - N = number of images
      - M = number of points for that image
      - rvecs = list of 1x3 Rodrigues rotation parameters
      - tvecs = list of 3x1 translation vectors
      - camera_matrix = [3x3] ndarray containing fx, fy, cx, cy
      - dist_coeffs = [1x5] ndarray, containing distortion coefficients

    :param object_points: Vector (N) of Vector (M) of 1x3 points of type float
    :param image_points: Vector (N) of Vector (M) of 1x2 points of type float
    :param image_size: (x, y) tuple, size in pixels, e.g. (1920, 1080)
    :param camera_matrix: Initial camera intrinsic matrix
    :param distortion_coefficients: Initial camera distortion coefficients
    :param flags: OpenCV flags to pass to calibrateCamera().
    :return: RMS reprojection error, camera_matrix, dist_coeffs, rvecs, tvecs
    """
    if image_size[0] < 1:
        raise ValueError("Image width must be > 0.")
    if image_size[1] < 1:
        raise ValueError("Image height must be > 0.")
    if len(object_points) < 2:
        raise ValueError("Must have at least 2 sets of object points.")
    if len(image_points) < 2:
        raise ValueError("Must have at least 2 sets of image points.")
    if len(object_points) != len(image_points):
        raise ValueError("Image points and object points differ in length.")
    for i, _ in enumerate(object_points):
        if len(object_points[i]) < 3:
            raise ValueError(str(i) + ": Must have at least 3 object points.")
        if len(image_points[i]) < 3:
            raise ValueError(str(i) + ": Must have at least 3 image points.")
        if len(object_points[i]) != len(image_points[i]):
            raise ValueError(str(i) + ": Must have the same number of points.")
    if camera_matrix is not None:
        cvm.validate_camera_matrix(camera_matrix)

    rms, camera_matrix, dist_coeffs, rvecs, tvecs \
        = cv2.calibrateCamera(object_points,
                              image_points,
                              image_size,
                              cameraMatrix=camera_matrix,
                              distCoeffs=distortion_coefficients,
                              flags=flags)

    return rms, camera_matrix, dist_coeffs, rvecs, tvecs


# pylint:disable=too-many-arguments,too-many-statements,too-many-branches
def mono_handeye_calibration(object_points: List,
                             image_points: List,
                             camera_matrix: np.ndarray,
                             camera_distortion: np.ndarray,
                             device_tracking_array: List,
                             pattern_tracking_array: List,
                             rvecs: List[np.ndarray],
                             tvecs: List[np.ndarray],
                             override_pattern2marker: np.ndarray = None,
                             use_opencv = True):
    """
    Wrapper around handeye calibration functions and reprojection /
    reconstruction error metrics.

    :param object_points: Vector of Vectors of 1x3 object points, float32
    :param image_points: Vector of Vectors of 1x2 object points, float32
    :param camera_matrix: Camera intrinsic matrix
    :param camera_distortion: Camera distortion coefficients
    :param device_tracking_array: Tracking data for camera (hand)
    :param pattern_tracking_array: Tracking data for calibration target
    :param rvecs: Vector of [3x1] ndarray, Rodrigues rotations for each camera
    :param tvecs: Vector of [3x1] ndarray, translations for each camera
    :param override_pattern2marker: If provided a 4x4 pattern2marker that
        is taken as constant.
    :param use_opencv: If false, use Guofang Xiao's method, otherwise OpenCV.
    :return: RMS reprojection error, handeye matrix, pattern to marker matrix
    """
    if len(object_points) < 3:
        raise ValueError("Must have at least 3 sets of object points.")
    if len(image_points) < 3:
        raise ValueError("Must have at least 3 sets of image points.")
    if len(object_points) != len(image_points):
        raise ValueError("Image points and object points differ in length.")
    # pylint: disable=too-many-boolean-expressions
    if rvecs is None or len(rvecs) == 0 or tvecs is None or len(tvecs) == 0 or \
        len(rvecs) != len(image_points) or len(tvecs) != len(object_points):
        # If we don't have all the rvecs and tvecs, its likely that the caller
        # has not run OpenCV calibrateCamera, but instead has provided
        # camera_matrix and camera_distortion. In this case, we can compute
        # rvec and tvec by using solvePnP.
        rvecs = []
        tvecs = []
        # pylint: disable=consider-using-enumerate
        for counter in range(len(object_points)):
            _, rvec, tvec = cv2.solvePnP(objectPoints=object_points[counter],
                                         imagePoints=image_points[counter],
                                         cameraMatrix=camera_matrix,
                                         distCoeffs=camera_distortion)
            rvecs.append(rvec)
            tvecs.append(tvec)
    has_device_tracking = False
    if device_tracking_array is not None and len(device_tracking_array) >= 3:
        for item in device_tracking_array:
            if item is None:
                raise ValueError("Device tracking array contains None item.")
        has_device_tracking = True
    has_pattern_tracking = False
    if pattern_tracking_array is not None and len(pattern_tracking_array) >= 3:
        for item in pattern_tracking_array:
            if item is None:
                raise ValueError("Pattern tracking array contains None item.")
        has_pattern_tracking = True
    if not has_device_tracking and not has_pattern_tracking:
        raise ValueError("There is insufficient device and/or tracking data.")
    if has_device_tracking and has_pattern_tracking:
        if len(device_tracking_array) != len(pattern_tracking_array):
            raise ValueError("Device tracking and camera tracking differ in length.")

    # Pad tracking lists with identity matrix if not provided.
    if has_device_tracking and not has_pattern_tracking:
        pattern_tracking_array = [np.eye(4) for _ in range(len(device_tracking_array))]
    if has_pattern_tracking and not has_device_tracking:
        device_tracking_array = [np.eye(4) for _ in range(len(pattern_tracking_array))]

    # Do main hand-eye calibration methods.
    if not use_opencv:

        # Guofang's method does both hand-eye and pattern-to-marker,
        # optimising both rotations quaternions then translation.

        quat_model2hand_array, trans_model2hand_array = \
            he.set_model2hand_arrays(pattern_tracking_array,
                                     device_tracking_array,
                                     use_quaternions=False)

        h2e, p2m =  \
            he.guofang_xiao_handeye_calibration(rvecs, tvecs,
                                                quat_model2hand_array,
                                                trans_model2hand_array)
    else:
        h2e, p2m = \
            he.calibrate_hand_eye_and_pattern_to_marker(
                rvecs,
                tvecs,
                device_tracking_array,
                pattern_tracking_array,
                method=cv2.CALIB_ROBOT_WORLD_HAND_EYE_SHAH
            )

    # We may have a pattern2marker override, i.e. you know it by design.
    if override_pattern2marker is not None:
        p2m = override_pattern2marker

    # Now, we optimise the final transformation.
    # Having read Ali et al (2019), https://doi.org/10.3390/s19122837
    # we should optimise projection error.

    if has_pattern_tracking and override_pattern2marker is None:
        # Now optimise p2m and h2e
        x_0 = np.zeros(12)
        rvec, tvec = vu.extrinsic_matrix_to_vecs(p2m)
        x_0[0] = rvec[0][0]
        x_0[1] = rvec[1][0]
        x_0[2] = rvec[2][0]
        x_0[3] = tvec[0][0]
        x_0[4] = tvec[1][0]
        x_0[5] = tvec[2][0]

        rvec, tvec = vu.extrinsic_matrix_to_vecs(h2e)
        x_0[6] = rvec[0][0]
        x_0[7] = rvec[1][0]
        x_0[8] = rvec[2][0]
        x_0[9] = tvec[0][0]
        x_0[10] = tvec[1][0]
        x_0[11] = tvec[2][0]

        res = minimize(vcf.mono_proj_err_p2m_h2e, x_0,
                       args=(object_points,
                             image_points,
                             camera_matrix,
                             camera_distortion,
                             pattern_tracking_array,
                             device_tracking_array
                             ),
                       method='Powell',
                       )

        x_1 = res.x
        rvec[0][0] = x_1[0]
        rvec[1][0] = x_1[1]
        rvec[2][0] = x_1[2]
        tvec[0][0] = x_1[3]
        tvec[1][0] = x_1[4]
        tvec[2][0] = x_1[5]
        p2m = vu.extrinsic_vecs_to_matrix(rvec, tvec)

        rvec[0][0] = x_1[6]
        rvec[1][0] = x_1[7]
        rvec[2][0] = x_1[8]
        tvec[0][0] = x_1[9]
        tvec[1][0] = x_1[10]
        tvec[2][0] = x_1[11]
        h2e = vu.extrinsic_vecs_to_matrix(rvec, tvec)

    else:
        # Now optimise just the h2e
        x_0 = np.zeros(6)
        rvec, tvec = vu.extrinsic_matrix_to_vecs(h2e)
        x_0[0] = rvec[0][0]
        x_0[1] = rvec[1][0]
        x_0[2] = rvec[2][0]
        x_0[3] = tvec[0][0]
        x_0[4] = tvec[1][0]
        x_0[5] = tvec[2][0]

        res = minimize(vcf.mono_proj_err_h2e, x_0,
                       args=(object_points,
                             image_points,
                             camera_matrix,
                             camera_distortion,
                             pattern_tracking_array,
                             device_tracking_array,
                             p2m
                             ),
                       method='Powell',
                       )

        x_1 = res.x
        rvec[0][0] = x_1[0]
        rvec[1][0] = x_1[1]
        rvec[2][0] = x_1[2]
        tvec[0][0] = x_1[3]
        tvec[1][0] = x_1[4]
        tvec[2][0] = x_1[5]
        h2e = vu.extrinsic_vecs_to_matrix(rvec, tvec)


    sse, num_samples = \
        vm.compute_mono_2d_err_handeye(object_points,
                                       image_points,
                                       camera_matrix,
                                       camera_distortion,
                                       device_tracking_array,
                                       pattern_tracking_array,
                                       h2e,
                                       p2m
                                       )

    mse = sse / num_samples
    reproj_err = np.sqrt(mse)

    return reproj_err, h2e, p2m


# pylint: disable=too-many-locals, too-many-arguments
def stereo_video_calibration(left_ids,
                             left_object_points,
                             left_image_points,
                             right_ids,
                             right_object_points,
                             right_image_points,
                             image_size,
                             left_intrinsics,
                             left_distortion,
                             right_intrinsics,
                             right_distortion,
                             mono_flags = 0,
                             stereo_flags = 0,
                             minimum_points = 10
                             ):
    """
    Default stereo calibration, using OpenCV methods.

    We wrap it here, so we have a place to add extra validation code,
    and a space for documentation. The aim is to check everything before
    we pass it to OpenCV, and raise Exceptions consistently for any error
    we can detect before we pass it to OpenCV.

    :param left_ids: Vector of ndarrays containing integer point ids.
    :param left_object_points: Vector of Vectors of 1x3 object points, float32
    :param left_image_points:  Vector of Vectors of 1x2 object points, float32
    :param right_ids: Vector of ndarrays containing integer point ids.
    :param right_object_points: Vector of Vectors of 1x3 object points, float32
    :param right_image_points: Vector of Vectors of 1x2 object points, float32
    :param image_size: (x, y) tuple, size in pixels, e.g. (1920, 1080)
    :param left_intrinsics: Initial camera intrinsic matrix
    :param left_distortion: Initial camera distortion coefficients
    :param right_intrinsics: Initial camera intrinsic matrix
    :param right_distortion: Initial camera distortion coefficients
    :param mono_flags: OpenCV flags to pass to calibrateCamera().
    :param stereo_flags: OpenCV flags to pass to stereoCalibrate().
    :param minimum_points: Minimum number of points that must match in left and right views.
    :return:
    """

    _, l_c, l_d, l_rvecs, l_tvecs \
        = cv2.calibrateCamera(left_object_points,
                              left_image_points,
                              image_size,
                              cameraMatrix=left_intrinsics,
                              distCoeffs=left_distortion,
                              flags=mono_flags
                              )

    _, r_c, r_d, r_rvecs, r_tvecs \
        = cv2.calibrateCamera(right_object_points,
                              right_image_points,
                              image_size,
                              cameraMatrix=right_intrinsics,
                              distCoeffs=right_distortion,
                              flags=mono_flags
                              )

    # For stereo, OpenCV needs common points.
    _, common_object_points, common_left_image_points, \
        common_right_image_points \
        = vu.filter_common_points_all_images(left_ids,
                                             left_object_points,
                                             left_image_points,
                                             right_ids,
                                             right_image_points, minimum_points=minimum_points)

    # This needs resetting, as the frames with common points may be fewer.
    number_of_frames = len(common_object_points)

    # Now, the stereo calib.
    _, l_c, l_d, r_c, r_d, \
        l2r_r, l2r_t, essential, fundamental = cv2.stereoCalibrate(
            common_object_points,
            common_left_image_points,
            common_right_image_points,
            l_c,
            l_d,
            r_c,
            r_d,
            image_size,
            flags=stereo_flags)

    # Normal OpenCV stereo calibration optimises intrinsics,
    # distortion, and stereo parameters, but doesn't output pose.
    # So here, we recompute the left camera pose.
    l_rvecs = []
    l_tvecs = []
    for i in range(0, number_of_frames):
        _, l_rvec, l_tvec = cv2.solvePnP(
            common_object_points[i],
            common_left_image_points[i],
            l_c,
            l_d)
        l_rvecs.append(l_rvec)
        l_tvecs.append(l_tvec)

    # Here, we are computing the right hand side rvecs and tvecs
    # given the new left hand side rvecs, tvecs and the l2r.
    left_to_right = skcm.construct_rigid_transformation(l2r_r, l2r_t)

    r_rvecs = []
    r_tvecs = []

    for i in range(0, number_of_frames):
        left_chessboard_to_camera = \
            vu.extrinsic_vecs_to_matrix(l_rvecs[i], l_tvecs[i])
        right_chessboard_to_camera = \
            np.matmul(left_to_right, left_chessboard_to_camera)
        r_rvec, r_tvec = \
            vu.extrinsic_matrix_to_vecs(right_chessboard_to_camera)
        r_rvecs.append(r_rvec)
        r_tvecs.append(r_tvec)

    # And recompute stereo projection error, given left camera and l2r.
    sse, num_samples = \
        vm.compute_stereo_2d_err(l2r_r,
                                 l2r_t,
                                 common_object_points,
                                 common_left_image_points,
                                 l_c,
                                 l_d,
                                 common_object_points,
                                 common_right_image_points,
                                 r_c,
                                 r_d,
                                 l_rvecs,
                                 l_tvecs
                                 )
    mse = sse / num_samples
    s_reproj = np.sqrt(mse)

    sse, num_samples = \
        vm.compute_stereo_3d_error(l2r_r,
                                   l2r_t,
                                   common_object_points,
                                   common_left_image_points,
                                   l_c,
                                   l_d,
                                   common_right_image_points,
                                   r_c,
                                   r_d,
                                   l_rvecs,
                                   l_tvecs
                                   )

    mse = sse / num_samples
    s_recon = np.sqrt(mse)

    LOGGER.info("Stereo Calib: proj=%s, recon=%s",
                str(s_reproj), str(s_recon))

    return s_reproj, s_recon, \
        l_c, l_d, l_rvecs, l_tvecs, \
        r_c, r_d, r_rvecs, r_tvecs, \
        l2r_r, l2r_t, \
        essential, fundamental


# pylint:disable=too-many-arguments,too-many-statements
def stereo_handeye_calibration(l2r_rmat: np.ndarray,
                               l2r_tvec: np.ndarray,
                               left_ids: List,
                               left_object_points: List,
                               left_image_points: List,
                               right_ids: List,
                               right_image_points: List,
                               left_camera_matrix: np.ndarray,
                               left_camera_distortion: np.ndarray,
                               right_camera_matrix: np.ndarray,
                               right_camera_distortion: np.ndarray,
                               device_tracking_array: List,
                               calibration_tracking_array: List,
                               left_rvecs: List[np.ndarray],
                               left_tvecs: List[np.ndarray],
                               override_pattern2marker=None,
                               use_opencv: bool = True,
                               minimum_points = 10
                               ):
    """
    Wrapper around handeye calibration functions and reprojection /
    reconstruction error metrics.

    :param l2r_rmat: [3x3] ndarray, rotation for l2r transform
    :param l2r_tvec: [3x1] ndarray, translation for l2r transform
    :param left_ids: Vector of ndarrays containing integer point ids.
    :param left_object_points: Vector of Vector of 1x3 of type float32
    :param left_image_points: Vector of Vector of 1x2 of type float32
    :param right_ids: Vector of ndarrays containing integer point ids.
    :param right_image_points: Vector of Vector of 1x3 of type float32
    :param left_camera_matrix: Camera intrinsic matrix
    :param left_camera_distortion: Camera distortion coefficients
    :param right_camera_matrix: Camera intrinsic matrix
    :param right_camera_distortion: Camera distortion coefficients
    :param device_tracking_array: Tracking data for camera (hand)
    :param calibration_tracking_array: Tracking data for calibration target
    :param left_rvecs: Vector of 3x1 ndarray, Rodrigues rotations for each
        camera
    :param left_tvecs: Vector of [3x1] ndarray, translations for each camera
    :param override_pattern2marker: If provided a 4x4 pattern2marker that
        is taken as constant.
    :param use_opencv: If True we use OpenCV based methods, if false,
        Guofang Xiao's method.
    :param minimum_points: Minimum number of points that must match in left and right views.
    :return: Reprojection error, reconstruction error, left handeye matrix,
        left pattern to marker matrix, right handeye, right pattern to marker
    :rtype: float, float, np.ndarray, np.ndarray, np.ndarray, np.ndarray
    """

    # First, we do mono calibration, for maximum code re-use.
    _, left_handeye_matrix, left_pattern2marker_matrix = \
        mono_handeye_calibration(
            left_object_points,
            left_image_points,
            left_camera_matrix,
            left_camera_distortion,
            device_tracking_array,
            calibration_tracking_array,
            left_rvecs,
            left_tvecs,
            override_pattern2marker=override_pattern2marker,
            use_opencv=use_opencv
        )

    # Filter common image points
    _, common_object_pts, common_l_image_pts, common_r_image_pts = \
        vu.filter_common_points_all_images(
            left_ids, left_object_points, left_image_points,
            right_ids, right_image_points,
            minimum_points)

    # Ensure right side is consistent.
    l2r_matrix = skcm.construct_rigid_transformation(l2r_rmat, l2r_tvec)
    right_handeye_matrix = l2r_matrix @ left_handeye_matrix
    right_pattern2marker_matrix = copy.deepcopy(left_pattern2marker_matrix)

    # Now compute some output statistics.
    if len(calibration_tracking_array) > 3 \
            and calibration_tracking_array[0] is not None:

        sse, num_samples = vm.compute_stereo_2d_err_handeye(
            common_object_pts,
            common_l_image_pts,
            left_camera_matrix,
            left_camera_distortion,
            common_r_image_pts,
            right_camera_matrix,
            right_camera_distortion,
            device_tracking_array,
            calibration_tracking_array,
            left_handeye_matrix,
            left_pattern2marker_matrix,
            right_handeye_matrix,
            right_pattern2marker_matrix
        )
        mse = sse / num_samples
        reproj_err = np.sqrt(mse)

        sse, num_samples = vm.compute_stereo_3d_err_handeye(
            l2r_rmat,
            l2r_tvec,
            common_object_pts,
            common_l_image_pts,
            left_camera_matrix,
            left_camera_distortion,
            common_r_image_pts,
            right_camera_matrix,
            right_camera_distortion,
            device_tracking_array,
            calibration_tracking_array,
            left_handeye_matrix,
            left_pattern2marker_matrix,
        )
        mse = sse / num_samples
        recon_err = np.sqrt(mse)

    else:

        sse, num_samples = vm.compute_stereo_2d_err(l2r_rmat,
                                                    l2r_tvec,
                                                    common_object_pts,
                                                    common_l_image_pts,
                                                    left_camera_matrix,
                                                    left_camera_distortion,
                                                    common_object_pts,
                                                    common_r_image_pts,
                                                    right_camera_matrix,
                                                    right_camera_distortion,
                                                    left_rvecs,
                                                    left_tvecs
                                                    )
        mse = sse / num_samples
        reproj_err = np.sqrt(mse)

        recon_err, num_samples = \
            vm.compute_stereo_3d_error(l2r_rmat,
                                       l2r_tvec,
                                       common_object_pts,
                                       common_l_image_pts,
                                       left_camera_matrix,
                                       left_camera_distortion,
                                       common_r_image_pts,
                                       right_camera_matrix,
                                       right_camera_distortion,
                                       left_rvecs,
                                       left_tvecs
                                       )
        mse = sse / num_samples
        recon_err = np.sqrt(mse)

    return reproj_err, recon_err, \
        left_handeye_matrix, left_pattern2marker_matrix, \
        right_handeye_matrix, right_pattern2marker_matrix


def stereo_calibration_extrinsics(common_object_points,
                                  common_left_image_points,
                                  common_right_image_points,
                                  l_rvecs,
                                  l_tvecs,
                                  override_left_intrinsics,
                                  override_left_distortion,
                                  override_right_intrinsics,
                                  override_right_distortion,
                                  override_l2r_rmat,
                                  override_l2r_tvec):
    """
    Simply re-optimises the extrinsic parameters.
    :return: error, l_rvecs, l_tvecs
    """
    number_of_frames = len(common_object_points)
    number_of_parameters = 6 * number_of_frames
    x_0 = np.zeros(number_of_parameters)
    for i in range(0, number_of_frames):
        x_0[i * 6 + 0] = l_rvecs[i][0][0]
        x_0[i * 6 + 1] = l_rvecs[i][1][0]
        x_0[i * 6 + 2] = l_rvecs[i][2][0]
        x_0[i * 6 + 3] = l_tvecs[i][0][0]
        x_0[i * 6 + 4] = l_tvecs[i][1][0]
        x_0[i * 6 + 5] = l_tvecs[i][2][0]

    res = least_squares(vcf.stereo_2d_error_for_extrinsics, x_0,
                        args=(common_object_points,
                              common_left_image_points,
                              common_right_image_points,
                              override_left_intrinsics,
                              override_left_distortion,
                              override_right_intrinsics,
                              override_right_distortion,
                              override_l2r_rmat,
                              override_l2r_tvec),
                        method='lm',
                        x_scale='jac',
                        verbose=0)

    LOGGER.info("Stereo Re-Calibration: status=%s", str(res.status))
    LOGGER.info("Stereo Re-Calibration: success=%s", str(res.success))
    LOGGER.info("Stereo Re-Calibration: msg=%s", str(res.message))

    x_1 = res.x
    for i in range(0, number_of_frames):
        l_rvecs[i][0][0] = x_1[i * 6 + 0]
        l_rvecs[i][1][0] = x_1[i * 6 + 1]
        l_rvecs[i][2][0] = x_1[i * 6 + 2]
        l_tvecs[i][0][0] = x_1[i * 6 + 3]
        l_tvecs[i][1][0] = x_1[i * 6 + 4]
        l_tvecs[i][2][0] = x_1[i * 6 + 5]

    return res.fun, l_rvecs, l_tvecs
